@@ -20,3 +20,8 @@ pool_metadata <- function(x,
   meta
 }
 
+# Lightweight equivalent of withr::defer()
+defer <- function(expr, envir = parent.frame(), after = FALSE) {
+  thunk <- as.call(list(function() expr))
+  do.call(on.exit, list(thunk, TRUE, after), envir = envir)
+}
